"""Load chat messages from common communications platforms for finetuning.

This module provides functions to load chat messages from various
communications platforms such as Facebook Messenger, Telegram, and
WhatsApp. The loaded chat messages can be used for finetuning models.
"""
